//=============================================================================
// UGHlink - TCP/IP textual communication for UGHstats mutator
// by }TCP{Wolf aka Mia_The_Chaotic
//
// change history
// build 2 (2006-08-07)
// - added trim() function
// - added isWhiteSpace() function
// build 3 (2006-08-25)
// - bug fix pointed out by Smartball: could not distinguish specs
//=============================================================================

class UGHlink extends TcpLink;

// references
var int port;			// port of stats server, set by UGHstats.uc
var UGHstats mut;		// reference back to the mutator to read out settings...


// runtime
var bool bOpened;		// set true if socket was opened successfully...
var bool bTerminated;	// set true when socket closed
var bool bPBG;			// Unreal engine bug... sometimes PostBeginPlay is called TWICE

var String data[5000];	// up to 5000 lines of text buffer which we may send...


// defaults
var int byteFragmentationSize;	// size of text-chunks, since Unreal cannot handle
							// large amounts of data in one piece



function PostBeginPlay()
{
	if (bPBG) return;	// ignore second post begin play call in case Unreal screws up again
	bPBG = true;

	// prepare the connection parameters, set to "text mode" and bind a local port
	// to use for transmission....
	LinkMode = MODE_Text;
	BindPort();
	super.PostBeginPlay();
}



// Called when domain resolution is successful.
// The IpAddr struct Addr contains the valid address.
event Resolved( IpAddr Addr )
{
	Addr.port = port;
	l("DNS/IP resolve successful... opening socket...");
	open(Addr);
	SetTimer(15, False);	// timeout in 10 seconds
}



// Called when domain resolution fails.
event ResolveFailed()
{
	l("FAILED TO RESOLVE STATS SERVER ADDRESS!! CHECK CONFIG!!");
	bTerminated = true;
	close();
}



function timer()
{
	if (bTerminated) return;	// make sure this is only called once or not at all if resolve failed
	bTerminated = true;

	if (!bOpened)
		l("Connection timed out!");
	/*else*/

	l("Terminating connection!");
	close();	// was open... end link
}



// called when connection was established successfully
event Opened()
{
	local int i;
	local int numFragments;
	local int ByteCounter;
	local String result, fragment;
	local UdpServerQuery udpquery;
	local TeamGameFix tgame;

	bOpened = True;

	// find udpquery
	foreach AllActors(class'UdpServerQuery', udpquery) break;
	if (udpquery == none)
	{
		l("CRITICAL ERROR, SERVER DOES CURRENTLY NOT HAVE A UDP QUERY ACTOR!");
		l("CANNOT SEND STATS!");
		// ABORT IMMEDIATELY!
		close();
		destroy();
		return;
	}

	// collect data
	result = "";
	if (mut.bReportBasic)
		result = result$udpquery.getBasic();
//log(len(result));
	if (mut.bReportInfo)
		result = result$udpquery.getInfo();
//log(len(result));
	if (mut.bReportRules)
		result = result$udpquery.getRules();
//log(len(result));

	// always report team game statistics in teamgames!
	if (mut.bgi.bTeamGame)
	{
		tgame = TeamGameFix(mut.bgi);
		result = result$"\\maxteams\\"$tgame.MaxTeams;		// maximum numbner of teams
		result = result$"\\maxteamsize\\"$tgame.MaxTeamSize;	// maximum size of a team
		for (i=0; i<4; i++)		// all teams...
		{
			result = result$"\\teamscore_"$i$"\\"$int(tgame.Teams[i].Score);	// team score of this team
			result = result$"\\teamsize_"$i$"\\"$tgame.Teams[i].Size;		// number of players in this team
			result = result$"\\teamname_"$i$"\\"$tgame.Teams[i].TeamName;	// name of this team (usually red, blue...)
		}
	}
//log(len(result));
	if (mut.bReportUTFadditions)
		result = result$CollectUTFadditions();
//log(len(result));
	if ((mut.bReportPlayers)||(mut.bReportBots))
		result = result$CollectPlayerStats();
//log(len(result));

	// we have collected all the data, now we need to split it up in packets
	// of a max size of 256, because Unreal will moan if we try to send a
	// large amount of text in one piece... so, for good measure we will
	// split the result text up in pieces of 250 bytes each (yea we're wasting
	// 6 bytes per array component who gives a damn...)

	numFragments = 0;	// we will use this variable to keep track how many text
					// fragments we got!

	while (len(result) > byteFragmentationSize)
	{
		// catch the next 250 bytes into "fragment"
		fragment = left(result,byteFragmentationSize);
		// cut away the left 250 bytes from the original...
		result = right(result,len(result)-byteFragmentationSize);
		// copy the fragment into the next data array holder unit
		data[numFragments++] = fragment;
	}
	// add the final piece
	data[numFragments++] = result;

	// Send stuff
	ByteCounter = 0;
	for (i=0; i<numFragments; i++)
		ByteCounter += (SendText(data[i]));

	l("Data transmission job fired for a total of "$ByteCounter$" bytes...");
	//close();	// may not close immediately, because this will terminate send...
}



event Closed()
{
	bTerminated = true;
	l("Connection was closed.");
	mut = none;
	destroy();
}



//////////////////////////////////////////////////////////////////////////////////
// PLAYER REPORT - original code from IpServer.UdpServerQuery
// modified:


function String CollectPlayerStats()
{
	local String result;
	local Pawn p;
	local PlayerReplicationInfoFIX pri;	// UTeamFix specific player replication info
	local bool bot, human, spec;
	local int PlayerNum;

	// go through all pawns
	PlayerNum = 0;						// counting starts at 0 or 1??
	for(p=Level.Pawnlist; p!=none; p=p.nextPawn)
	{
		if (p.bIsPlayer && (p.PlayerReplicationInfo!=none))		// player is participating visibly on the scoreboard
		{
			human = p.IsA('PlayerPawn');
			bot = p.IsA('UTeamBots');

			// spectator detection bugfix - reported by Smartball
			pri = PlayerReplicationInfoFIX(p.PlayerReplicationInfo);
			if (pri != none) spec = pri.bIsSpectator; // else false :roll:

			// add?
			if (!bot && !human)
				l("Now here's something odd: "$pri.PlayerName$" is a player, but neither a player nor bot - IGNORING!");
			else if ( (bot && mut.bReportBots) || (human && mut.bReportPlayers && (mut.bReportSpecs || !spec)) )
			{
				// name...
				result = result$"\\player_"$PlayerNum$"\\"$cleanPlayerName(pri.PlayerName);

				// admin?
				if (human)
					result = result$"\\admin_"$PlayerNum$"\\"$PlayerPawn(p).bAdmin;

				// spec?
				if (human)
					result = result$"\\spectator_"$PlayerNum$"\\"$spec;

				// type
				if (human) result = result$"\\type_"$PlayerNum$"\\human";
				if (bot)   result = result$"\\type_"$PlayerNum$"\\bot";

				// Score
				result = result$"\\score_"$PlayerNum$"\\"$int(pri.Score);

				// Frags
				result = result$"\\frags_"$PlayerNum$"\\"$pri.Frags;

				// Deaths
				result = result$"\\deaths_"$PlayerNum$"\\"$pri.Deaths;

				// Suicides
				result = result$"\\suicides_"$PlayerNum$"\\"$pri.Suicides;

				// Time (in seconds)
				result = result$"\\timeinseconds_"$PlayerNum$"\\"$pri.TimeSeconds;

				// Teamkills
				if (mut.bgi.bTeamGame || mut.bgi.griFIX.bCoop)	// report teamkills only for TEAM and COOP games
					result = result$"\\teamkills_"$PlayerNum$"\\"$pri.teamkills;

				// Team number (what team is the playe on 0-3)
				if (mut.bgi.bTeamGame)	// no report for DeathMatch (team is always 255 then)
					result = result$"\\team_"$PlayerNum$"\\"$pri.Team;

				// Ping
				if (bot)
					result = result$"\\ping_"$PlayerNum$"\\0"; // build 2: killed whitespace
				else
					result = result$"\\ping_"$PlayerNum$"\\"$trim(P.ConsoleCommand("GETPING")); // build 2: killed whitespace
			}
		} // end "player is participating..."
		PlayerNum++;
	} // end for pawnlist

 return result;
}


// collect a few pieces of information I find useful myself
// but you can always turn it off
function String CollectUTFadditions()
{
	local String ResultSet;

	// map FILENAME - some maps have no internal name or one containing
	// ascii chars, but there always is a normal filename
	ResultSet = ResultSet$"\\mapfilename\\"$mut.bgi.griFIX.CurrentMap;

	// standoff stats?
	if (mut.bgi.StandOffMode.bStandOff)
	{
		ResultSet = ResultSet$"\\roundlimit\\"$mut.bgi.StandOffMode.RoundLimit;
		ResultSet = ResultSet$"\\roundtimelimit\\"$mut.bgi.StandOffMode.RoundTimeLimit;
		ResultSet = ResultSet$"\\standoffscoringrule\\"$mut.bgi.Standoff_ExtraKillPoints;
	}

	return ResultSet;
}

	// TODO ADD
	// MM?



///////////////////////////////////////////////
// AID

// clean a name from ascii chars
function String cleanPlayerName(String pname)
{
 local String result, check;
 local  int i;

 result = "";
 
 for (i=0; i<len(pname); i++)
 {
	check = mid(pname,i,1);
	if (isValidChar_PlayerName(check))
		result = result $ check;
	else
		result = result $ "."; // dot is a replacement character
 }
 return result;
}


// this check basically sorts out all the "non-displayable" ascii chars
function bool isValidChar_PlayerName(String s)
{
 local int ivalue;

 ivalue = ASC(s);
 if (ivalue == 33) return true; // !
 if (ivalue == 92) return false;
 if ( (ivalue >= 35) && (ivalue <= 125) ) return true; // # to }
 return false;
}


// new function since build 2
// function cleans out white-space characters at beginning and end of a string
function String trim(String s)
{
 local String result;
 result = s;

//l("to trim>>>"$result$"<<<");
 // sanity check
 if (len(s)<1) return s;

 while (isWhiteSpace(left(result,1)))
 	result = right(result,len(result)-1);
 while (isWhiteSpace(right(result,1)))
 	result = left(result,len(result)-1);

//l("to resu>>>"$result$"<<<");
 return result;
}


// new function since build 2
// determine exactly WHAT a whitespace is... s here should be len=1
function bool isWhiteSpace(String s)
{
 if (s == " ") return true;
 if (!isValidChar_PlayerName(s)) return true;
 return false;
}




// log function... I am such a lazy writer :P
function l(String what)
{
 log(what,'UGHlink');
}



defaultproperties
{
	byteFragmentationSize=250
}
